# This program is free software: you can redistribute it and/| modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, |
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY | FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.
from django.core.management.base import BaseCommand
from django.utils.translation import gettext as _
from list_based_flavorpack.utils import clone_list_based_pool

class Command(BaseCommand):
    help = _('Clones a list-based pool by machine name')

    def add_arguments(self, parser):
        parser.add_argument('machine_name', type=str, help='The machine name of the pool to be cloned')
        parser.add_argument('new_machine_name', type=str, help='The new machine name of the pool to be cloned')
    
    def handle(self, *args, **options):
        print(_('Clones a list-based pool by machine name'))
        clone_list_based_pool(options['machine_name'], options['new_machine_name'])



# Generated by Django 3.0.2 on 2020-04-01 21:45

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('list_based_flavorpack', '0007_auto_20191219_1759'),
    ]

    operations = [
        migrations.AlterField(
            model_name='listbasedregion',
            name='processing_class_path',
            field=models.CharField(default='list_based_flavorpack.processing_classes.third_party_processing.processing.DBProcessingClass', help_text='The full python path to the class that will be processing region allocations', max_length=150, verbose_name='Processing Class Path'),
        ),
    ]



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.
import os
import uuid
import time
import sqlite3
import requests
from lxml import etree
from urllib.parse import urlparse
from django.utils.translation import gettext as _
from requests.auth import HTTPBasicAuth, HTTPProxyAuth
from quartet_capture.rules import RuleContext
from quartet_output import errors
from quartet_output.transport.http import HttpTransportMixin, user_agent
from quartet_output.models import EndPoint
from quartet_capture import models, rules, errors as capture_errors
from list_based_flavorpack.models import ListBasedRegion
from list_based_flavorpack.processing_classes.third_party_processing.rules import \
    get_region_table


class NumberRequestTransportStep(rules.Step, HttpTransportMixin):
    '''
    Uses the transport information within the `region`.
    '''

    def execute(self, data, rule_context: RuleContext):
        # get the task parameters that we rely on
        try:
            self.info(
                _('Looking for the task parameter with the target Region. '
                  'Output Name.'))
            param = models.TaskParameter.objects.get(
                task__name=rule_context.task_name,
                name='List-based Region'
            )
            # now see if we can get the output critieria based on the param
            # value
            self.info(_('Found the region param, now looking up the '
                        'Region instance with name %s.'),
                      param.value
                      )
        except models.TaskParameter.DoesNotExist:
            raise capture_errors.ExpectedTaskParameterError(
                _('The task parameter with name List-based Region '
                  'could not be found.  This task parameter is required by '
                  'the NumberRequestTransportStep to function correctly.')
            )
        try:
            region = ListBasedRegion.objects.get(machine_name=param.value)
            # check the url/urn to see if we support the protocol
            protocol = self._supports_protocol(region.end_point)
            self.info('Protocol supported.  Sending message to %s.' %
                      region.end_point.urn)
            response = self._send_message(data, protocol, rule_context, region)
            if 'ErrorCode' in response.content.decode('utf-8'):
                raise self.ResponseError(response.content.decode('utf-8'))
            # Pass response for downstream processing.
            rule_context.context['NUMBER_RESPONSE'] = response.content
        except Exception as e:
            self.error(_(
                "An error occurred while sending request to third party: %s"),
                       str(e))
            raise

    class ResponseError(Exception):
        pass

    def get_auth(self, region):
        """
        Get's the authentication method and credentials from the
        region record.
        :param region: A ListBasedModel model instance.
        :return: A `requests.auth.HTTPBasicAuth` or `HTTPProxyAuth`
        """
        auth_info = region.authentication_info
        auth = None
        if auth_info:
            auth_type = auth_info.type or ''
            if 'digest' in auth_type.lower():
                auth = HTTPBasicAuth
            elif 'proxy' in auth_type.lower():
                auth = HTTPProxyAuth
            else:
                auth = HTTPBasicAuth
            auth = auth(auth_info.username, auth_info.password)
        return auth

    def post_data(self, data: str, rule_context: RuleContext,
                  region: ListBasedRegion,
                  content_type='application/xml',
                  file_extension='xml',
                  http_put=False):
        '''
        :param data_context_key: The key within the rule_context that contains
         the data to post.  If being invoked from the internals of this
         module this is usually the OUTBOUND_EPCIS_MESSSAGE_KEY value of the
         `quartet_output.steps.ContextKeys` Enum.
        :param output_criteria: The output criteria containing the connection
        info.
        :return: The response.
        '''
        if not http_put:
            func = requests.post
        else:
            func = requests.put
        response = func(
            region.end_point.urn,
            data,
            auth=self.get_auth(region),
            headers={'content-type': content_type, 'user-agent': user_agent}
        )
        return response

    def _send_message(
            self,
            data: str,
            protocol: str,
            rule_context: RuleContext,
            region: ListBasedRegion
    ):
        '''
        Sends a message using the protocol specified.
        :param protocol: The scheme of the urn in the output_criteria endpoint.
        :param rule_context: The RuleContext contains the data in the
        OUTBOUND_EPCIS_MESSAGE_KEY value from the `ContextKey` class.
        :param region: The originating region.
        :return: None.
        '''
        content_type = self.get_parameter('content-type', 'application/xml')
        file_extension = self.get_parameter('file-extension', 'xml')
        put_data = self.get_boolean_parameter('put-data')
        if protocol.lower() in ['http', 'https']:
            if not put_data:
                func = self.post_data
            else:
                func = self.put_data
            self.info("request sent %s", data)
            response = func(
                data,
                rule_context,
                region,
                content_type,
                file_extension
            )
            try:
                response.raise_for_status()
            except:
                if response.content:
                    self.info("Error occurred with following response %s",
                              response.content)
                raise
            self.info("Response Received %s", response.content)
            return response

    def _supports_protocol(self, endpoint: EndPoint):
        '''
        Inspects the output settings and determines if this step can support
        the protocol or not. Override this to support another or more
        protocols.
        :param EndPoint: the endpoint to inspect
        :return: Returns the supported scheme if the protocol is supported or
        None.
        '''
        parse_result = urlparse(
            endpoint.urn
        )
        if parse_result.scheme.lower() in ['http', 'https']:
            return parse_result.scheme
        else:
            raise errors.ProtocolNotSupportedError(_(
                'The protocol specified in urn %s is not supported by this '
                'step or module.'
            ), endpoint.urn)

    def on_failure(self):
        super().on_failure()

    @property
    def declared_parameters(self):
        return {
            'content-type': 'The content-type to add to the header during any '
                            'http posts, puts, etc. Default is application/'
                            'xml',
            'file-extension': 'The file extension to specify when posting and '
                              'putting data via http. Default is xml'
        }


class UUIDRequestStep(rules.Step):
    '''
    Returns UUIDs.
    Excepts a request data as follows:
    <?xml version="1.0" encoding="UTF-8"?>
    <NumberRequest>
        <Type>UUID</Type>
        <Size>10</Size>
    </NumberRequest>
    '''

    def execute(self, data, rule_context: RuleContext):
        # get the task parameters that we rely on
        try:
            self.info(
                _('Looking for the task parameter with the target Region. '
                  'Output Name.'))
            param = models.TaskParameter.objects.get(
                task__name=rule_context.task_name,
                name='List-based Region'
            )
            # now see if we can get the output critieria based on the param
            # value
            self.info(_('Found the region param, now looking up the '
                        'Region instance with name %s.'),
                      param.value
                      )
        except models.TaskParameter.DoesNotExist:
            raise capture_errors.ExpectedTaskParameterError(
                _('The task parameter with name List-based Region '
                  'could not be found.  This task parameter is required by '
                  'the NumberRequestTransportStep to function correctly.')
            )
        try:
            region = ListBasedRegion.objects.get(machine_name=param.value)
            # check the url/urn to see if we support the protocol
            root = etree.fromstring(data)
            size = int(root.find('.//Size').text)
            self.persist_data(region, size)
        except Exception as e:
            self.info(_(
                "An error occurred while sending request to third party: %s"),
                      str(e))
            self.info(_("Inbound data %s"), str(data))
            raise

    def persist_data(self, region, size):
        """
        Saves the data to file.  Override to save it other places.
        :param region: The region to save the data for.
        :param size: The number of UUIDs to generate.
        :return: None
        """
        with open(region.file_path, "a") as f:
            for i in range(size):
                f.write("%s\n" % uuid.uuid1())

    def on_failure(self):
        super().on_failure()

    @property
    def declared_parameters(self):
        return {
            'content-type': 'The content-type to add to the header during any '
                            'http posts, puts, etc. Default is application/'
                            'xml',
            'file-extension': 'The file extension to specify when posting and '
                              'putting data via http. Default is xml'
        }


class UUIDRequestDBStep(UUIDRequestStep):
    """
    Instead of saving data to file will save to a sqlite3 database file.
    """

    def persist_data(self, region, size):
        start = time.time()
        connection = sqlite3.connect(region.db_file_path)
        cursor = connection.cursor()
        cursor.execute('begin transaction')
        self.info('storing the numbers.')
        for i in range(size):
            cursor.execute('insert into %s (serial_number, used) '
                           'values (?, ?)' % get_region_table(region),
                           (str(uuid.uuid1()), 0))
        cursor.execute('commit')
        self.info("Execution time: %.3f seconds." % (time.time() - start))



'''
    Copyright 2018 SerialLab, CORP

    This file is part of ListBasedFlavorpack.

    RandomFlavorpack is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    RandomFlavorpack is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with RandomFlavorpack.  If not, see <http://www.gnu.org/licenses/>.
'''

from django.conf.urls import url
from list_based_flavorpack.api import views
from list_based_flavorpack.api import viewsets
from list_based_flavorpack.api.routers import urlpatterns as router_urlpatterns

urlpatterns = [
    url(r'^list-based-regions/$',
        viewsets.list_based_region_list,
        name='list-based-region-list'),
    url(r'^list-based-region-create/$',
        viewsets.list_based_region_create,
        name='list-based-region-create'),
    url(r'^list-based-region-detail/(?P<machine_name>[\w\-\_]{1,100})/$',
        viewsets.list_based_region_detail,
        name='list-based-region-detail'),
    url(r'^list-based-region-modify/(?P<machine_name>[\w\-\_]{1,100})/$',
        viewsets.list_based_region_modify,
        name='list-based-region-modify'),
    url(r'^list-based-region-form/(?P<machine_name>[\w\-\_]{1,100})/$',
        viewsets.list_based_region_form,
        name='list-based-region-form'),
        url(r'^clone-list-based-pool/$',
        views.CloneListBasedFlavorpackView.as_view(),
        name='clone-pool'),
    url(
        r'^clone-list-based-pool/(?P<machine_name>[\w\-\_]{1,100})/(?P<new_machine_name>[\w\-\_]{1,100})/$',
        views.CloneListBasedFlavorpackView.as_view(),
        name='clone-pool-view')
] + router_urlpatterns



'''
    Copyright 2018 SerialLab, CORP

    This file is part of ListBasedFlavorpack.

    ListBasedFlavorpack is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    ListBasedFlavorpack is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with ListBasedFlavorpack. If not, see <http://www.gnu.org/licenses/>.
'''
from serialbox.generators.common import Generator
from serialbox.rules.common import PreprocessingRule
from list_based_flavorpack import list_based_flavorpack_settings as settings

def import_processing_class(region):
    '''
    Returns the processing class dynamically.
    A processing class is entered when creating a list-based region
    and is used for all rules/generator logic.
    '''
    package_path = region.processing_class_path.split(".")
    class_path = package_path[:-1]
    class_name = package_path[-1]
    processing_module = __import__(".".join(class_path), fromlist=[class_name])
    return getattr(processing_module, class_name)


class ListBasedGenerator(Generator):
    '''
    Generates a list of numbers (or anything else located in source)
    '''
    def generate(self, request, response, region, size):
        '''
        Executes an instance of the processing class and returns a list.
        '''
        processing_class = import_processing_class(region)
        return self.set_number_list(response, processing_class().get_items(region, size))

    def get_settings_module(self):
        return settings



class ListBasedPreprocessRule(PreprocessingRule):
    '''
    Used as a proxy for all rules that are added to the Processing Class.
    This allows for dynamic injection of rules without having to hack
    settings too much (except for adding this rule)
    '''
    def execute(self, request, pool, region, size):
        '''
        This will execute all the rules that have been added to
        the processing class (which is entered dynamically when creating
        a list-based region.)
        '''
        processing_instance = import_processing_class(region)()
        for rule in processing_instance.get_pre_processing_rules():
            rule().execute(request, pool, region, size)


class ListBasedPostprocessRule(PreprocessingRule):
    '''
    Used as a proxy for all rules that are added to the Processing Class.
    This allows for dynamic injection of rules without having to hack
    settings too much (except for adding this rule)
    '''
    def execute(self, request, pool, region, size):
        '''
        This will execute all the rules that have been added to
        the processing class (which is entered dynamically when creating
        a list-based region.)
        '''
        processing_instance = import_processing_class(region)()
        for rule in processing_instance.get_post_processing_rules():
            rule().execute(request, pool, region, size)



'''
    Copyright 2018 SerialLab, CORP

    This file is part of ListBasedFlavorpack.

    ListBasedFlavorpack is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    ListBasedFlavorpack is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with ListBasedFlavorpack.  If not, see <http://www.gnu.org/licenses/>.
'''

from django.conf import settings


GENERATOR_PREPROCESSING_RULES = getattr(
    settings,
    'GENERATOR_PREPROCESSING_RULES',
    {
        'list_based_flavorpack.generators.list_based.ListBasedGenerator': [
            'serialbox.rules.limits.ActiveRule',
            'list_based_flavorpack.generators.list_based.ListBasedPreprocessRule',
            'serialbox.rules.limits.RequestThresholdLimitRule',
        ]
    })

GENERATOR_POSTPROCESSING_RULES = getattr(
    settings,
    'GENERATOR_POSTPROCESSING_RULES',
    {'list_based_flavorpack.generators.list_based.ListBasedGenerator': [
            'serialbox.rules.limits.ActiveRule',
            'list_based_flavorpack.generators.list_based.ListBasedPostprocessRule',
            'serialbox.rules.limits.RequestThresholdLimitRule',
        ]})
